from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


def get(hub, tree, ref: str):
    result = tree
    if ref:
        for key in ref.split("."):
            try:
                result = result[key]
            except KeyError:
                if (
                    "functions" in result
                    and "variables" in result
                    and "attributes" in result
                ):
                    if key in result["functions"]:
                        result = result["functions"][key]
                    elif key in result["variables"]:
                        result = result["variables"][key]
                    elif key in result.get("classes", {}):
                        result = result["classes"][key]
                else:
                    raise

        return {ref: result}
    else:
        return tree


def list_(hub, tree: Dict[str, Any]) -> Dict[str, Any]:
    """
    Return all the references available on the hub by reference first
    """
    ret = {}

    def _get_refs(t: Dict[str, Any]):
        for k, v in t.items():
            if isinstance(v, Dict):
                _get_refs(v)
            elif k == "ref":
                ret[t["ref"]] = t

    _get_refs(tree)

    return ret



from typing import Any
from typing import Dict
from typing import List

from dict_tools import data as data_


def load_all(hub, pypaths: List[str] = None):
    """
    Load the named pypaths and all available dynes onto the hub
    """
    if pypaths:
        for pypath in pypaths:
            try:
                hub.pop.sub.add(
                    pypath=pypath,
                    omit_class=False,
                    omit_func=False,
                    omit_vars=False,
                    stop_on_failures=False,
                )
            except ModuleNotFoundError as e:
                hub.log.error(e)
                raise

    for dyne in hub._dynamic:
        if not hasattr(hub, dyne):
            hub.pop.sub.add(
                dyne_name=dyne,
                omit_class=False,
                omit_func=False,
                omit_vars=False,
                stop_on_failures=False,
            )
        try:
            hub.pop.sub.load_subdirs(hub[dyne], recurse=True)
        except (AttributeError or TypeError):
            ...


async def recurse(hub, sub, ref: str = None) -> Dict[str, Any]:
    """
    Find all of the loaded subs in a Sub. I.E:
        pprint(hub.pop.tree.recurse(hub.pop))
    :param hub: The redistributed pop central hub
    :param sub: The pop object that contains the loaded module data
    :param ref: The current reference on the hub
    """
    sub_name = sub._dyne_name
    if sub_name:
        if ref:
            ref = f"{ref}.{sub_name}"
        else:
            ref = sub_name
    ret = data_.NamespaceDict()
    for loaded in sorted(sub._subs):
        loaded_ref = f"{ref}.{loaded}"
        try:
            loaded_sub = getattr(sub, loaded)
        except AttributeError:
            continue
        if not (
            getattr(loaded_sub, "_virtual", False)
            and getattr(loaded_sub, "_sub_virtual", True)
        ):
            # Bail early if the sub's virtual isn't True
            continue
        recursed_sub = await hub.tree.sub.recurse(loaded_sub, ref=loaded_ref)

        for mod in sorted(loaded_sub._loaded):
            loaded_mod = getattr(loaded_sub, mod)
            recursed_sub[mod] = await hub.tree.mod.parse(
                loaded_mod, ref=f"{ref}.{loaded}.{mod}"
            )

        if recursed_sub:
            ret[loaded] = recursed_sub

    return ret



CLI_CONFIG = {
    "output": {
        "source": "rend",
        "default": None,
    },
    "ref": {
        "positional": True,
        "nargs": "?",
    },
    "recurse": {
        "action": "store_true",
    },
    "graph": {},
    "graph_layout": {},
    "hide": {"nargs": "*"},
    "pypaths": {"nargs": "*"},
}

CONFIG = {
    "ref": {
        "type": str,
        "help": "The ref on the hub to show",
        "default": None,
    },
    "pypaths": {
        "help": "A space delimited list of pypaths to add to the hub",
        "default": [],
    },
    "graph": {
        "help": "Plugin to use for generating a graph, (I.E. 'simple', 'details', 'json')",
        "default": None,
    },
    "graph_layout": {
        "help": "They layout to use with the graph plugin",
        "default": None,
    },
    "hide": {
        "help": "Hide teh named node types from the graph. I.e. functions, contracts, variables",
        "default": [],
    },
}

DYNE = {
    "graph": ["graph"],
    "tree": ["tree"],
}



from typing import Any
from typing import Dict
from typing import List

__virtualname__ = "details"


def __virtual__(hub):
    return (
        hasattr(hub, "output"),
        "'rend' needs to be installed in the python environment to use this plugin",
    )


def process_mod(
    hub,
    ref: str,
    doc: str,
    file: str,
    attributes: List[str],
    functions: Dict[str, Dict[str, Any]],
    variables: Dict[str, Dict[str, Any]],
    classes: Dict[str, Dict[str, Any]],
):
    ...


def show(hub, tree: Dict[str, Any]):
    outputter = hub.OPT.rend.output or "nested"
    rendered = hub.output[outputter].display(tree)
    print(rendered)



from typing import Any
from typing import Dict

from dict_tools import data as data_

__func_alias__ = {"format_": "format"}


def __init__(hub):
    hub.pop.sub.add(dyne_name="rend", omit_class=False)
    hub.pop.sub.add(dyne_name="graph", omit_class=False)


def doc_cli(hub):
    hub.pop.config.load(["pop_doc", "pop_tree", "rend"], cli="pop_doc")
    hub.pop.loop.create()
    hub.tree.sub.load_all()

    tree = hub.pop.Loop.run_until_complete(hub.tree.init.traverse())
    tree = hub.tree.ref.get(tree, hub.OPT.pop_doc.ref)
    ret = hub.tree.ref.list(tree)

    if not (ret or tree):
        raise KeyError(f"Reference does not exist on the hub: {hub.OPT.pop_doc.ref}")

    result = ret[hub.OPT.pop_doc.ref]

    outputter = hub.OPT.rend.output or "nested"
    print(hub.output[outputter].display(result))


def cli(hub):
    hub.pop.config.load(["pop_tree", "rend"], cli="pop_tree")
    hub.pop.loop.create()
    hub.tree.sub.load_all(pypaths=hub.OPT.pop_tree.pypaths)

    tree = hub.pop.Loop.run_until_complete(hub.tree.init.traverse())
    result = hub.tree.ref.get(tree, hub.OPT.pop_tree.ref)

    if hub.OPT.pop_tree.graph:
        hub.graph.GRAPH = hub.OPT.pop_tree.graph
    else:
        # Find the first plugin that was loaded for graphing
        loaded_mods = hub.graph._loaded
        if "simple" in loaded_mods:
            hub.graph.GRAPH = "simple"
        else:
            iter_mods = iter(hub.graph._loaded)
            hub.graph.GRAPH = next(iter_mods)
            if hub.graph.GRAPH == "init":
                hub.graph.GRAPH = next(iter_mods)

    hub.graph.init.show(result)


async def traverse(hub) -> Dict[str, Any]:
    """
    :param hub: The redistributed pop central hub
    :return: A dictionary representation of all the subs on the hub. I.E:
        pprint(hub.pop.tree.traverse())
    """
    root = data_.NamespaceDict()
    for sub in hub._iter_subs:
        loaded_sub = getattr(hub, sub)
        root[sub] = await hub.tree.sub.recurse(loaded_sub)

        for loaded_mod in sorted(loaded_sub, key=lambda x: x.__name__):
            mod = loaded_mod.__name__
            root[sub][mod] = await hub.tree.mod.parse(loaded_mod, ref=f"{sub}.{mod}")
    return root



import inspect
import re
import textwrap
from collections import OrderedDict
from dataclasses import is_dataclass
from typing import _GenericAlias
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

import aiofiles

UNKNOWN_REF = "< unknown ref >"


def get_ref(hub, mod) -> str:
    """
    Try to find a reference on the hub for the given mod
    """
    try:
        sister_func = next(iter(mod._funcs.values()))
        return sister_func.ref
    except StopIteration:
        ...


def serialize_signature(
    hub, signature: inspect.Signature, param_aliases: Dict[str, Set[str]]
):
    ret = _serialize_params(signature.parameters, param_aliases)

    final_ret = {"parameters": ret}
    if signature.return_annotation is not inspect._empty:
        final_ret["return_annotation"] = signature.return_annotation
    return final_ret


def _serialize_params(params, aliases: Dict[str, Set[str]] = None):
    if aliases is None:
        aliases = {}
    ret = OrderedDict()
    for p in params:

        param: inspect.Parameter = params[p]
        ret[param.name] = {}
        if p in aliases:
            ret[param.name]["aliases"] = sorted(aliases[p])

        if param.default is not inspect._empty:
            ret[param.name]["default"] = param.default

        annotation = param.annotation
        if annotation is inspect._empty:
            continue

        # If there are aliases for this parameter, then the annotation might be the first value in a tuple
        if p in aliases:
            if isinstance(annotation, str):
                # The annotation is just a string containing an "alias=..."
                continue
            else:
                try:
                    # See if the annotation is a tuple (<actual annotation>, "alias=", ...)
                    # The first argument can be the actual annotation, the rest are string aliases
                    annotation = annotation[0]
                except IndexError:
                    # The annotation is
                    continue

        # Check if the annotation is "Computed", "Sensitive", or any other type created from "typing.Optional"
        match = re.match(r"typing\.Union\[(.*)\]", str(annotation))
        if match:
            try:
                # The sub_type is the second value from the Union
                ret[param.name]["sub_type"] = annotation.__args__[1]
                # The real annotation is the first value from the Union
                annotation = annotation.__args__[0]
            except:
                ...

        param_type = annotation
        if isinstance(param_type, _GenericAlias) and annotation.__origin__ == list:
            param_type = annotation.__args__[0]

        if is_dataclass(param_type):
            # Serialize parameters of the dataclass init method
            ret[param.name]["annotation"] = {
                # Serialize parameters of the Dataclass init method
                # Skip self argument of the Dataclass init method
                str(annotation): _serialize_params(
                    {
                        k: v
                        for k, v in inspect.signature(
                            param_type.__init__
                        ).parameters.items()
                        if k != "self"
                    },
                )
            }
        else:
            ret[param.name]["annotation"] = str(annotation)

    return ret


def format_func(hub, f: Callable, **kwargs):
    lines, start_line = hub.tree.mod.get_source_lines(f)
    return {
        "doc": textwrap.dedent(str(f.__doc__ or "")).strip("\n"),
        "file": inspect.getfile(f),
        "start_line_number": start_line,
        "end_line_number": start_line + len(lines),
        **kwargs,
    }


def funcs(hub, mod, ref: str) -> List[str] or Dict[str, str]:
    """
    Find all of the loaded functions in a pop plugin. I.E:
        pprint(hub.pop.tree.funcs(hub.pop.tree))
    :param hub: The redistributed pop central hub
    :param mod: A plugin that has been loaded onto a sub
    :param ref: The current reference on the hub
    :return: A Dictionary of loaded modules names mapped to a list of their functions
    """
    funcs = sorted(mod._funcs.keys())
    ret = {}
    for f in funcs:
        contract = mod._funcs[f]
        func_info = hub.tree.mod.format_func(
            contract.func,
            ref=f"{ref}.{f}",
            contracts={
                contract_type: [f"{c.ref}.{c.func.__name__}" for c in contracts]
                for contract_type, contracts in contract.contract_functions.items()
            },
        )
        func_info.update(
            hub.tree.mod.serialize_signature(
                contract.signature, contract.param_aliases
            ),
        )
        ret[f] = func_info
    return ret


def _format_var(name: str, value: Any, source_lines: List[str], **kwargs):
    line_number = 0
    for num, line in enumerate(source_lines):
        if name in line:
            line_number = num + 1
            break

    return {
        "type": value.__class__.__name__,
        "value": value,
        "start_line_number": line_number,
        **kwargs,
    }


async def data(hub, mod, ref: str) -> List[str] or Dict[str, str]:
    """
    Find all of the loaded data in a pop plugin. I.E:
        pprint(hub.pop.tree.data(hub.pop.tree))
    :param hub: The redistributed pop central hub
    :param mod: A plugin that has been loaded onto a sub
    :param ref: The current reference on the hub
    """
    datas = sorted(x for x in mod._vars if x.isupper() and not x.startswith("_"))
    ret = {}

    source_file = inspect.getsourcefile(mod)
    async with aiofiles.open(source_file, "r") as fh:
        lines = await fh.readlines()

    for d_name in datas:
        ret[d_name] = _format_var(
            d_name, mod._vars[d_name], lines, ref=f"{ref}.{d_name}", file=source_file
        )

    return ret


async def types(hub, mod, ref: str) -> List[str] or Dict[str, str]:
    """
    Find all of the loaded types in a pop plugin. I.E:
        pprint(hub.pop.tree.types(hub.pop.tree))
    :param hub: The redistributed pop central hub
    :param mod: A plugin that has been loaded onto a sub
    :param ref: The current reference on the hub
    """
    classes = sorted(x for x in mod._classes if not x.startswith("_"))
    ret = {}
    for class_name in classes:
        c = mod._classes[class_name]
        source_file = inspect.getsourcefile(c)

        async with aiofiles.open(source_file, "r") as fh:
            source_lines = await fh.readlines()

        try:
            lines, start_line = hub.tree.mod.get_source_lines(c)
        except OSError:
            start_line = 0
            for num, line in enumerate(source_lines):
                if class_name in line:
                    start_line = num + 1
                    break
            lines = []

        signature = inspect.signature(c.__init__)
        functions = {}
        variables = {}
        attributes = []
        for name, value in inspect.getmembers(c):
            if name.startswith("_"):
                continue
            attributes.append(name)
            attr_ref = f"{ref}.{class_name}.{name}"
            if inspect.isfunction(value):
                functions[name] = hub.tree.mod.format_func(value, ref=attr_ref)
            else:
                variables[name] = _format_var(
                    name, value, source_lines, ref=attr_ref, file=source_file
                )
        class_info = {
            "ref": f"{ref}.{class_name}",
            "doc": textwrap.dedent(c.__doc__ or "").strip("\n"),
            "signature": hub.tree.mod.serialize_signature(signature, {}),
            "attributes": attributes,
            "functions": functions,
            "variables": variables,
            "file": source_file,
            "start_line_number": start_line,
            "end_line_number": start_line + len(lines),
        }
        ret[class_name] = class_info
    return ret


async def parse(hub, mod, ref: str) -> Dict[str, Any]:
    """
    Parse a loaded mod object

    :param hub: The redistributed pop central hub
    :param mod: A plugin that has been loaded onto a sub
    :param ref: The current reference on the hub
    """
    return {
        "ref": ref,
        "doc": (mod._attrs.get("__doc__") or "").strip(),
        "file": getattr(mod, "__file__", None),
        "attributes": sorted(
            a for a in mod._attrs if not (a.startswith("__") and a.endswith("__"))
        ),
        "classes": await hub.tree.mod.types(mod, ref),
        "functions": hub.tree.mod.funcs(mod, ref),
        "variables": await hub.tree.mod.data(mod, ref),
    }


def get_source_lines(hub, module: Callable) -> Tuple[List, int]:
    try:
        return inspect.getsourcelines(module)
    except OSError:
        # This can happen when running pop-tree from within a binary
        return [], -1



from dataclasses import field
from dataclasses import make_dataclass
from typing import List

from dict_tools.typing import Computed


def str_alias(hub, param: "alias=p"):
    return param


def multiple_str_alias(hub, param: "alias=p, alias=prm"):
    return param


def tuple_alias(hub, param: (str, "alias=p")):
    return param


def multiple_tuple_alias(hub, param: (str, "alias=p", "alias=prm")):
    return param


def tuple_alias_dataclass(
    hub,
    param: (
        List[
            make_dataclass(
                "Tag",
                [
                    ("Key", str, field(default=None)),
                    ("Value", str, field(default=None)),
                ],
            )
        ],
        "alias=p",
        "alias=prm",
    ),
):
    return param


def computed(hub, param: Computed[str] = None):
    return param


def computed_nest(
    hub,
    param: Computed[
        List[
            make_dataclass(
                "Tag",
                [
                    ("Key", str, field(default=None)),
                    ("Value", str, field(default=None)),
                ],
            )
        ]
    ],
):
    return param


def computed_alias(hub, param: (Computed[str], "alias=p")):
    return param



#!/usr/bin/env python3
import os
import pathlib
import shutil

from setuptools import Command
from setuptools import setup

DESC = "Visualize POP hub/sub structures"

VERSION = "10.1.0"

SETUP_DIRNAME = os.path.dirname(__file__)
if not SETUP_DIRNAME:
    SETUP_DIRNAME = os.getcwd()

with open("README.rst", encoding="utf-8") as f:
    LONG_DESC = f.read()

with open(os.path.join("requirements", "base.txt")) as f:
    REQUIREMENTS = f.read().splitlines()

REQUIREMENTS_EXTRA = {"full": set()}
EXTRA_PATH = pathlib.Path("requirements", "extra")
for extra in EXTRA_PATH.iterdir():
    with extra.open("r") as f:
        REQUIREMENTS_EXTRA[extra.stem] = f.read().splitlines()
        REQUIREMENTS_EXTRA["full"].update(REQUIREMENTS_EXTRA[extra.stem])


class Clean(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for subdir in ("pop_create", "pop_docs", "tests"):
            for root, dirs, files in os.walk(
                os.path.join(os.path.dirname(__file__), subdir)
            ):
                for dir_ in dirs:
                    if dir_ == "__pycache__":
                        shutil.rmtree(os.path.join(root, dir_))


def discover_packages():
    modules = []
    for package in ("pop_tree", "pop_doc"):
        for root, _, files in os.walk(os.path.join(SETUP_DIRNAME, package)):
            pdir = os.path.relpath(root, SETUP_DIRNAME)
            modname = pdir.replace(os.sep, ".")
            modules.append(modname)
    return modules


setup(
    name="pop-tree",
    author="VMware, Inc.",
    author_email="idemproject@vmware.com",
    url="https://vmware.gitlab.io/pop/pop-tree/en/latest/index.html",
    project_urls={
        "Code": "https://gitlab.com/vmware/pop/pop-tree",
        "Issue tracker": "https://gitlab.com/vmware/pop/pop-tree/issues",
    },
    version=VERSION,
    install_requires=REQUIREMENTS,
    extras_require=REQUIREMENTS_EXTRA,
    description=DESC,
    long_description=LONG_DESC,
    long_description_content_type="text/x-rst",
    python_requires=">=3.7",
    license="Apache Software License 2.0",
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
    ],
    packages=discover_packages(),
    entry_points={
        "console_scripts": [
            "pop-tree = pop_tree.scripts:start",
            "pop-doc = pop_doc.scripts:start",
        ]
    },
    cmdclass={"clean": Clean},
)
